#include <Servo.h>
#include <TimedAction.h>
#include <SimpleTimer.h>           // For movement tests
#include <Wire.h>
#include <LCDi2cR.h>

const int base =       0;          // 60; 160 ccw, 0 cw
const int shoulder =   1;          // 130, 180 (full back); 120 (forward)
const int elbow =      2;          // 110, 150 fwd, 60 back (full back when shoulder fwd)
const int wristflex =  3;          // 20, 160 full back, 20 full down
const int wristrot =   4;          // 110, 10 ccw, 180 cw
const int gripper =    5;          // 95, 160 closed, 20 open

const int stdDelay =   20;
const int maxServos =  6;
const int centerPos =  90;

unsigned long key_millis = 0;
unsigned long button_millis = 0;
int keyDelay = 100;                    // 500 non-calibrate
int buttonDelay = 50;
int thisServo = base;

typedef struct{
  byte neutral;
  byte minPos;
  byte maxPos;
  byte delaySpeed;
  byte curPos;
} ServoPos;

ServoPos servosPos[] = {
  { 60, 160, 0, stdDelay, 0 },
  { 130, 180, 120, stdDelay, 0 },
  { 110, 150, 60, stdDelay, 0 },
  { 20, 160, 5, stdDelay, 0 },
  { 110, 180, 10, 10, 0 },
  { 95, 160, 20, 5, 0 }
};

byte serv = 90;
int counter = 0;
int curServo = 0;
int sMove[] = {0, 90, 0};
int sAttach[] = {0, 0};

LCDi2cR lcd = LCDi2cR(4,40,0x63,0); 
Servo servos[maxServos];
int destServoPos[maxServos];
int currentServoPos[maxServos];

TimedAction servoMove[maxServos] = TimedAction(100, doServoFunc);   // do delay as const
SimpleTimer timer;            // For movement tests
TimedAction keys = TimedAction(10, keypadFunc);

void setup() {
  delay(200);
  Wire.begin();
  lcd.init(); 
  delay(500);
  lcd.on();
  setupDisplay();

  for(int i=0; i<maxServos; i++) {
    servos[i].write(servosPos[i].neutral);
    servosPos[i].curPos = servosPos[i].neutral;
    servos[i].attach(i+4);
    destServoPos[i] = centerPos;
    currentServoPos[i] = centerPos;
    servoMove[i].disable();
  }
  //timer.setInterval(5000, servoTestFunc);      // For movement tests
}

void loop() {
  for(int x=0; x<maxServos; x++) {
    curServo = x;
    servoMove[x].check();
  }
  //timer.run();            // For movement tests
  keys.check();
  navSwitchFunc();
}

void servoTestFunc() {
  if(counter % 2) {
    Move(thisServo, servosPos[thisServo].minPos, servosPos[thisServo].delaySpeed);
  }
  else
    Move(thisServo, servosPos[thisServo].maxPos, servosPos[thisServo].delaySpeed);
  counter++;
}

void writeServo() {
  int servoNum = sMove[0];
  if(servoNum >=0 && servoNum <= maxServos) {
    destServoPos[servoNum] = sMove[1];
    servoMove[servoNum].enable();
    servoMove[servoNum].setInterval(sMove[2]);
  }
}

void setServoAttach() {
  int servo = 1;    // sAttach[0]
  int mode = 2;     // sAttach[1]
  if(servo >= 0 && servo <= maxServos) {
    if (mode == 1)
      servos[servo].attach(servo+4);
    else
      servos[servo].detach();
  }
}

void doServoFunc() {
  int x = curServo;      // don't use global here; keep x local
 
  if(destServoPos[x] == currentServoPos[x])
    servoMove[x].disable();    // move to end?
    
  if(destServoPos[x] > currentServoPos[x])
    currentServoPos[x]++;
  else
    currentServoPos[x]--;
  
  servosPos[x].curPos = constrain(currentServoPos[x], servosPos[x].maxPos, servosPos[x].minPos);
  currentServoPos[x] = servosPos[x].curPos;
  servos[x].write(currentServoPos[x]);
  jointPos(x, currentServoPos[x]);
}

void Move(int servoNum, int servoPosition, int delayTime) {
  sMove[0] = servoNum;
  sMove[1] = servoPosition;
  sMove[2] = delayTime;
  writeServo();
}

void Attach(int servoNum, int servoMode) {
  sAttach[0] = servoNum;
  sAttach[1] = servoMode;
}

void navSwitchFunc() {
    if (millis() > button_millis + buttonDelay) {
    button_millis = millis();
    if(digitalRead(A0) == LOW) {      // white
      thisServo--;
      thisServo = constrain(thisServo, 0, 5);
      jointPos(thisServo, servosPos[thisServo].curPos);
      delay(200);
    }
    if(digitalRead(A1)== LOW) {      // orange
      thisServo++;
      thisServo = constrain(thisServo, 0, 5);
      jointPos(thisServo, servosPos[thisServo].curPos);
      delay(200);
    }
    if(digitalRead(A2) == LOW) {      // gray
      byte t = thisServo;
      servosPos[t].curPos--;
      servosPos[t].curPos = constrain(servosPos[t].curPos, servosPos[t].maxPos, servosPos[t].minPos);
      jointPos(t, servosPos[t].curPos);
    }
    if(digitalRead(A3) == LOW) {     // purple
      byte t = thisServo;
      servosPos[t].curPos++;
      servosPos[t].curPos = constrain(servosPos[t].curPos, servosPos[t].maxPos, servosPos[t].minPos);
      jointPos(t, servosPos[t].curPos);
    }
  }
}

void keypadFunc() {
  byte keypad = lcd.keypad();
  lcd.command(0);                // Send null; apparent bug fix
  if (keypad !=0) {
    if (millis() > key_millis + keyDelay) {
      keypress(keypad);
      key_millis = millis();
    }
  }
}

void keypress (byte keypad) {
  byte t;
  lcd.setCursor(0, 5);
  switch(keypad) {
    case 1: 
      lcd.print("1");
      t = base;
      servosPos[t].curPos--;
      break;
    case 2:
      lcd.print("2");
      t = shoulder;
      servosPos[t].curPos--;
      break;
    case 3:
      lcd.print("3");
      t = elbow;
      servosPos[t].curPos--;
      break;
    case 5:
      lcd.print("4");
      t = base;
      servosPos[t].curPos++;
      break;
    case 6:
      lcd.print("5");
      t = shoulder;
      servosPos[t].curPos++;
      break;
    case 7:
      lcd.print("6");
      t = elbow;
      servosPos[t].curPos++;
      break;
    case 9:
      lcd.print("7");
      t = wristflex;
      servosPos[t].curPos--;
      break;
    case 10:
      lcd.print("8");
      t = wristrot;
      servosPos[t].curPos--;
      break;
    case 11:
      lcd.print("9");
      t = gripper;
      servosPos[t].curPos--;
      break;
    case 13:
      lcd.print("*");
      t = wristflex;
      servosPos[t].curPos++;
      break;
    case 14:
      lcd.print("0");
      t = wristrot;
      servosPos[t].curPos++;
      break;
    case 15:
      lcd.print("#");
      t = gripper;
      servosPos[t].curPos++;
      break;   
  }
  servosPos[t].curPos = constrain(servosPos[t].curPos, servosPos[t].maxPos, servosPos[t].minPos);
  jointPos(t, servosPos[t].curPos);
}

void jointPos(byte t, byte pos) {
  lcd.setCursor(1, 7);
  switch(t) {
    case base:               // 1, 4
      lcd.print("bse");
      break;
    case shoulder:           // 2, 5     
      lcd.print("shl");
      break;
    case elbow:              // 3, 6
      lcd.print("elb");
      break;
    case wristflex:          // 7, * 
      lcd.print("wfx");
      break;
    case wristrot:           // 8, 0    
      lcd.print("wrt");
      break;
    case gripper:            // 9, #
      lcd.print("grp");
      break;
  }
  
  lcd.setCursor(1, 17);
  lcd.print("   ");
  lcd.setCursor(1, 17);
  lcd.print(pos, DEC);
  servos[t].write(pos);
}

void setupDisplay() {
  lcd.clear();
  lcd.blink_off();
  lcd.home();                                       // Cursor to home position (0,0)
  lcd.print("key: ");
  lcd.setCursor(1, 0);
  lcd.print("Joint:      ");
  lcd.setCursor(1, 12);
  lcd.print("Pos: ");
}
